function varargout = fy_ivac(cmd,job)
% Execution file for iVAC
% based on Christian Gaser's cg_tom8.m 7 2009-09-30 14:25:20Z gaser
% Fumio Yamashita
% fy_ivac.m 7 2012-07-25 14:15:00

linfun = inline('fprintf([''%-40s%s''],x,[repmat(sprintf(''\b''),1,40)])','x');

modi = {'Image'};

if strcmpi(cmd,'create')
    Image   = job.images;
    data = {Image};
    maskfile = job.mask{1};
    odir = job.odir{1};
    % for mean and standard deviation approach
    if isfield(job.database,'meansd')
        files = [];
        for j = 1:length(data{1})
            files = strvcat(files,data{1}{j});
        end
        fy_ivac_calcmeansd(files,odir,maskfile);
        % return output, if requested
        if nargout > 0
            varargout{1} = [];
        end
        return;
    end
    % for regression approach
    % initialise output argument
    out = struct('ivacmat',{{''}},'betas',{{''}},'betacovs',{{''}},'sigmas',{{''}});
    age     = job.database.regression.age.c;
    gender  = job.database.regression.gender.c;
    agepoly = fieldnames(job.database.regression.poly);
    covs = job.database.regression.cov;
    n_covs  = length(covs);
    
    switch agepoly{1}
        case 'poly1', ageDegree = 1;
        case 'poly2', ageDegree = 2;
        case 'poly3', ageDegree = 3;
    end
    
    if ~strcmp(data{1},'')
        n_modi = length(data{1});
    else
        n_modi = 0;
    end
    
    % check for correct number of covariates
    if n_modi~=size(age,1), error('Number of age values differ from number of images.'); end
    if n_modi~=size(gender,1), error('Number of gender values differ from number of images.'); end
    for i = 1:n_covs
        if n_modi~=size(covs(i).c,1)
            error(sprintf('Number of %s differ from number of images.',covs(i).cname));
        end
    end
    
    if (min(gender)<0) || (max(gender)>1)
        error('Valid values are 1 for males and 0 for females.');
    end
    
    % age regression with polynomial function
    X = zeros(n_modi(1),ageDegree+1);
    %for j = 0:ageDegree, X(:,(j + 1)) = (age.^j) - X*(pinv(X)*(age.^j)); end
    for j = 0:ageDegree, X(:,(j + 1)) = age.^j; end
    
    % calculate correction for orthogonalization
    for j = 1:ageDegree, ageReg{j} = pinv(X(:,1:j))*(age.^j); end
    
    %-Ask about overwriting files from previous analyses...
    %-------------------------------------------------------------------
    if exist(fullfile(odir,'iVAC.mat'),'file')
        str = {	'Current directory contains existing iVAC.mat file:',...
            'Continuing will overwrite all files!'};
        if spm_input(str,1,'bd','stop|continue',[1,0],1,mfilename);
            fprintf('%-40s: %30s\n\n',...
                'Abort...   (existing iVAC.mat file)',spm('time'));
            return
        end
    end
    
    % remove old files
    files = {'^beta\..{3}$','^betacov\..{3}$','^iVAC\..{3}$'};
    
    for i=1:length(files)
        j = spm_select('List',odir,files{i});
        for k=1:size(j,1)
            spm_unlink(fullfile(odir,deblank(j(k,:))));
        end
    end
    
    % extend design matrix with gender
    X = [X gender];
    for i = 1:n_covs, X = [X covs(i).c]; end
    
    n_beta = size(X,2);
    
    % save regression for orthogonalization
    out.ivacmat = {fullfile(odir,'iVAC.mat')};
    save(out.ivacmat{1},'-V6', 'X', 'ageReg', 'age', 'ageDegree', 'gender', 'covs', 'n_modi');
    
    pKX = pinv(X);
    W = zeros(n_modi(1));
    for i = 1:n_modi(1), W(i,i) = 1; end
    W = sparse(W);
    
    TH = -Inf*ones(n_modi(1),1);
    
    % names of covariates
    for i = 1:ageDegree+1, betanames{i} = ['age' num2str(i-1)]; end
    betanames{i+1} = 'gender';
    for j = 1:n_covs, betanames{j+i+1} = covs(j).cname; end
    
    % go through all modi which are defined
    for i = find(n_modi>0)
        
        fprintf('Estimating %s...\n',modi{i});
        
        P = [];
        for j = 1:length(data{i})
            P = strvcat(P,data{i}{j});
        end
        
        V = spm_vol(P);
        Vmask = spm_vol(maskfile);
        
        % estimate betas
        [beta,betacov,sigma] = calc_beta(V,X,pKX,Vmask); %% modified
        
        dim = V(1).dim(1:3);
        
        VO = V(1);
        VO.dt = [spm_type('float32') spm_platform('bigend')];
        
        beta = reshape(beta,[dim n_beta]);
        betacov = reshape(betacov,[dim n_beta^2]); %% added
        %sigma = reshape(sigma,dim);
        
        % save beta images with prepending mod-name
        out.betas = {fullfile(odir,'beta.nii')};
        out.betacovs = {fullfile(odir,'betacov.nii')};
        out.sigma = {fullfile(odir,'sigma.nii')};
        
        %% write out beta multivolume image
        Beta = nifti;
        Beta.dat = file_array(fullfile(odir,'beta.nii'),size(beta),[spm_type('float32') spm_platform('bigend')]);
        Beta.mat = VO.mat;
        Beta.mat0 = VO.mat;
        Beta.descrip = 'beta image';
        create(Beta);
        Beta.dat(:) = beta(:);
        %% write out beta covariance multivolume image
        BetaCov = nifti;
        BetaCov.dat = file_array(fullfile(odir,'betacov.nii'),size(betacov),[spm_type('float32') spm_platform('bigend')]);
        BetaCov.mat = VO.mat;
        BetaCov.mat0 = VO.mat;
        BetaCov.descrip = 'beta covariance image';
        create(BetaCov);
        BetaCov.dat(:) = betacov(:);
        %% write out sigma image
        Sigma = nifti;
        Sigma.dat = file_array(fullfile(odir,'sigma.nii'),size(sigma),[spm_type('float32') spm_platform('bigend')]);
        Sigma.mat = VO.mat;
        Sigma.mat0 = VO.mat;
        Sigma.descrip = 'sigma image';
        create(Sigma);
        Sigma.dat(:) = sigma(:);
        % return output, if requested
        if nargout > 0
            varargout{1} = out;
        end
        return;
    end
    
    %% individual analysis
elseif strcmpi(cmd,'analysis')
    % initialise output argument
    out = struct('timage',{{''}},'zimage',{{''}});
    Image   = job.images;
    maskfile = job.mask{1};
    data = {Image};
    odir = job.odir{1};
    % load and reslice mask image
    if ~isempty(maskfile)
        mask = reslice_mask(spm_vol(Image{1}),spm_vol(maskfile));
    end
    % for mean and standard deviation approach
    if isfield(job.database,'meansd')
        volmean = spm_vol(job.database.meansd.meanimage{1});
        volsd = spm_vol(job.database.meansd.sdimage{1});
        Meanimage = spm_read_vols(volmean);
        SDimage = spm_read_vols(volsd);
        for i=1:length(Image)
            inputimage = Image{i};
            [path,name,ext] = spm_fileparts(inputimage);
            if isempty(odir)
                outputdir = path;
            else
                outputdir = odir;
            end
            %% load input image
            VY = spm_vol(inputimage);
            Y = spm_read_vols(VY);
            zscoreimage = -(Y-Meanimage)./SDimage;% Waring! sign of z-score is inverted to represent atrophy by positive values
            if ~isempty(maskfile)
                zscoreimage = zscoreimage.*mask;
            end
            Vz = VY;
            Vz.fname = fullfile(outputdir,['z_' name ext]);
            Vz.dt = [spm_type('float32') spm_platform('bigend')];
            Vz.descrip = 'z-score image';
            Vz.pinfo(1) = 1;
            spm_write_vol(Vz,zscoreimage);
            out.zimage{i} = Vz.fname;
            spm_progress_bar('Set',i);
        end
        %% clear progress bar
        spm_progress_bar('Clear');
        
        % return output, if requested
        if nargout > 0
            varargout{1} = out;
        end
        return;
    end
    % for regression approach
    load(job.database.regression.ivacmat{:});
    betaimage = job.database.regression.betaimages{1};
    [path,name,ext] = spm_fileparts(betaimage);
    betaimage = fullfile(path,[name ext]);
    %betacovimage = job.database.regression.betacovimages{1};
    %[path,name,ext] = spm_fileparts(betacovimage);
    %betacovimage = fullfile(path,[name ext]);
    sigmaimage = job.database.regression.sigmaimage{1};
    [path,name,ext] = spm_fileparts(sigmaimage);
    sigmaimage = fullfile(path,[name ext]);
    age     = job.database.regression.age.c;
    gender  = job.database.regression.gender.c;
    outputimage = job.database.regression.outputimage;
    covs = job.database.regression.cov;
    n_covs  = length(covs);
    
    if ~strcmp(data{1},'')
        n_modi = length(data{1});
    else
        n_modi = 0;
    end
    
    % check for correct number of covariates
    if n_modi~=size(age,1), error('Number of age values differ from number of images.'); end
    if n_modi~=size(gender,1), error('Number of gender values differ from number of images.'); end
    for i = 1:n_covs
        if n_modi~=size(covs(i).c,1)
            error('Number of %s differ from number of images.',covs(i).cname);
        end
    end
    
    if (min(gender)<0) || (max(gender)>1)
        error('Valid values are 1 for males and 0 for females.');
    end
    
    % age regression with polynomial function
    X0s = zeros(n_modi(1),ageDegree+1);
    %for j = 0:ageDegree, X0(:,(j + 1)) = (age.^j) - X0*(pinv(X0)*(age.^j)); end
    for j = 0:ageDegree, X0s(:,(j + 1)) = age.^j; end
    % calculate correction for orthogonalization
    for j = 1:ageDegree, ageReg{j} = pinv(X0s(:,1:j))*(age.^j); end
    % extend design matrix with gender and covariates
    X0s = [X0s gender];
    for i = 1:n_covs, X0s = [X0s covs(i).c]; end
    
    %% load beta and sigma images
    V_beta = spm_vol(betaimage);
    %V_betaCov = spm_vol(betacovimage);
    beta = spm_read_vols(V_beta);
    %betaCov = spm_read_vols(V_betaCov);
    V_sigma = spm_vol(sigmaimage);
    sigma = spm_read_vols(V_sigma);
    tmpX = X'*X;
    tmpXinv = inv(tmpX);
    tmpX = tmpX(:);
    tmpXinv = tmpXinv(:);
    
    %% loop for t- and z- score calculation
    for i=1:length(Image)
        inputimage = Image{i};
        [path,name,ext] = spm_fileparts(inputimage);
        if isempty(odir)
            outputdir = path;
        else
            outputdir = odir;
        end
        %% load input image
        VY = spm_vol(inputimage);
        Y = spm_read_vols(VY);
        %% set X0
        X0 = X0s(i,:);
        %% estimate Yhat
        Yhat = beta(:,:,:,1);
        for j = 2:size(X0,2)
            Yhat = Yhat + beta(:,:,:,j)*X0(j);
        end
        %% calculate t- and SD image
        tmpX0 = X0'*X0;
        tmpX0 = tmpX0(:);
        if outputimage == 0
            timage = zeros(size(Y));
        elseif outputimage == 1
            SDimage = zeros(size(Y));
        else
            timage = zeros(size(Y));
            SDimage = zeros(size(Y));
        end
        betaCovslice = zeros(prod(V_beta(1).dim(1:2)),size(X0,2)^2);
        betaslice = zeros(prod(V_beta(1).dim(1:2)),size(X0,2));
        sigmaslice = zeros(prod(V_beta(1).dim(1:2)),1);
        for j=1:VY.dim(3)
            for k=1:size(X,2)
                tmp = beta(:,:,j,k);
                tmp(isnan(tmp)) = 0;
                betaslice(:,k) = tmp(:);
            end
            %             for k=1:size(X0,2)^2
            %                 tmp = betaCov(:,:,j,k);
            %                 tmp(isnan(tmp)) = 0;
            %                 betaCovslice(:,k) = tmp(:);
            %             end
            tmp = sigma(:,:,j);
            tmp(isnan(tmp)) = 0;
            sigmaslice = tmp(:);
            for k=1:size(X0,2)^2
                tmp = tmpXinv(k).*sigmaslice;
                tmp(isnan(tmp)) = 0;
                betaCovslice(:,k) = tmp(:);
            end
            if outputimage == 0 || outputimage == 2
                t = Yhat(:,:,j)./reshape(sqrt(betaCovslice*tmpX0),VY.dim(1:2));
                timage(:,:,j) = t;
            end
            if outputimage == 1 || outputimage == 2
                %sigma = betaCovslice*tmpX/size(X0,2);
                %SD = sqrt(betaCovslice*tmpX0+sigma);
                SD = sqrt(betaCovslice*tmpX0+sigmaslice);
                %SD = sigmaslice.*sqrt(1+X0*inv(X'*X)*X0');
                SD = reshape(SD,VY.dim(1:2));
                SDimage(:,:,j) = SD;
            end
            spm_progress_bar('Set',j);
        end
        %% write t-image
        if outputimage == 0 || outputimage == 2
            if ~isempty(maskfile)
                timage = timage.*mask;
            end
            Vt = VY;
            Vt.fname = fullfile(outputdir,['t_' name ext]);
            Vt.dt = [spm_type('float32') spm_platform('bigend')];
            Vt.descrip = 't-image';
            Vt.pinfo(1) = 1;
            spm_write_vol(Vt,timage);
            out.timage{i} = Vt.fname;
        end
        %%  write z-score image
        if outputimage == 1 || outputimage == 2
            zscoreimage = -(Y-Yhat)./SDimage;% Waring! sign of z-score is inverted to represent atrophy by positive values
            if ~isempty(maskfile)
                zscoreimage = zscoreimage.*mask;
            end
            Vz = VY;
            Vz.fname = fullfile(outputdir,['z_' name ext]);
            Vz.dt = [spm_type('float32') spm_platform('bigend')];
            Vz.descrip = 'z-score image';
            Vz.pinfo(1) = 1;
            spm_write_vol(Vz,zscoreimage);
            out.zimage{i} = Vz.fname;
        end
    end
    %% clear progress bar
    spm_progress_bar('Clear');
    
    % return output, if requested
    if nargout > 0
        varargout{1} = out;
    end
    return;
    %% roi analysis
elseif strcmpi(cmd,'roianalysis')
    % initialise output argument
    out = struct('roiresult',{{''}});
    Image   = job.images;
    maskfile = job.mask{1};
    roifile = job.roi{1};
    odir = job.odir{1};
    % load and reslice mask image
    if ~isempty(maskfile)
        mask = reslice_mask(spm_vol(Image{1}),spm_vol(maskfile));
    end
    % load and reslice roi image
    roi = reslice_voi(spm_vol(Image{1}),spm_vol(roifile));
    labels = unique(roi(:));
    labels = labels(labels~=0); % omit background
    %% save ROI result
    fp = fopen(fullfile(odir,['roiresult_' datestr(now,'yyyymmdd_HHMMSS') '.tsv']),'wt');
    fprintf(fp,'No\tImagefiles\t');
    for i=1:size(labels,1)-1
        fprintf(fp,['label' num2str(labels(i)) '\t']);
    end
    fprintf(fp,['label' num2str(labels(end)) '\n']);
    for i=1:size(Image,1)
        inputimage = Image{i};
        %% load input image
        VY = spm_vol(inputimage);
        Y = spm_read_vols(VY);
        if ~isempty(maskfile)
            roi = roi.*mask;
        end
        fprintf(fp,'%d\t%s\t',i,VY.fname);
        for j=1:size(labels,1)
            Ylabeled = Y(roi==labels(j));
            roiresult(i,j).mean = mean(Ylabeled(:));
            if j ~= size(labels,1)
                fprintf(fp,'%f\t',roiresult(i,j).mean);
            else
                fprintf(fp,'%f\n',roiresult(i,j).mean);
            end
        end
        out.roiresult{i} = roiresult(i,j);
        spm_progress_bar('Set',i);
    end
    fclose(fp);
    %% clear progress bar
    spm_progress_bar('Clear');
    
    % return output, if requested
    if nargout > 0
        varargout{1} = out;
    end
    return;
    
elseif strcmpi(cmd,'render')
    % initialise output argument
    out = struct('report',{{''}});
    Image   = job.images;
    renderfile = job.renderfile{1};
    templatefile = job.templatefile{1};
    range = job.range.c;
    odir = job.odir{1};
    outputformat = fieldnames(job.format);
    outputformat = outputformat{1};
    for i=1:length(Image)
        inputimage = Image{i};
        [path,name,ext] = spm_fileparts(inputimage);
        if isempty(odir)
            outputdir = path;
        else
            outputdir = odir;
        end
        %% load input image
        vol = spm_vol(inputimage);
        parametricimage = spm_read_vols(vol);
        parametricimage(isinf(parametricimage)) = NaN;
        parametricimage = parametricimage.*(parametricimage>=range(1)&parametricimage<=range(2));
        parametricimage2d = parametricimage(parametricimage>0);
        %%% render %%%
        [row,col] = find(parametricimage>0);
        indy = rem(col-1,vol.dim(2))+1;
        indz = floor(col/vol.dim(2))+1;
        dat = struct('XYZ', [row(:) indy(:) indz(:)]',...
            't',parametricimage2d',...
            'mat',vol.mat,...
            'dim',vol.dim);
        fy_spm_render(dat,NaN,renderfile);
        %%% draw slices using slover %%%
        sloverobj = draw_slices(inputimage,templatefile,range);
        fh = sloverobj.figure;
        figure(fh);
        %set(fh, 'PaperPositionMode', 'auto');
        set(fh,'PaperPosition',[0.03 0 0.94 1]);
        ax = axes;
        set(ax,'Position',[0 0.92 1 0.07],'Visible','off','XDir','normal','YDir','reverse');
        text(0,0,path,'Color',[0.5 0.5 0.5],'fontsize',8);
        text(0,0.4,[name ext],'fontsize',10);
        %%% print results  %%%
        outputfilename = fullfile(outputdir,[name '.' outputformat]);
        switch outputformat
            case 'png'
                print(fh,'-dpng',outputfilename,'-r150');
            case 'tiff'
                print(fh,'-dtiff',outputfilename,'-r150');
            case 'pdf'
                print(fh,'-dpdf',outputfilename,'-r150');
        end
        out.report = {outputfilename};
    end
    
elseif strcmpi(cmd,'getinfo')
    % dummy output
    out = [];
    
    load(job.ivacmat{:});
    
    fprintf('\nInformation about %s\n\n',job.ivacmat{:});
    % print range of age and gender
    fprintf('Age:\t%d values (%3.2f .. %3.2f years)\n',length(age),min(age),max(age));
    fprintf('Gender:\t%d females / %d males\n',sum(find(gender==0)>0),sum(find(gender==1)>0));
    
    % print information about covariates
    n_covs = length(covs);
    for i = 1:n_covs
        fprintf('Covariate %d:\t%s\n',i,covs(i).cname);
    end
    
    % print order
    fprintf('Order of age regression: %d\n',length(ageReg));
end

% return output, if requested
if nargout > 0
    varargout{1} = out;
end
return

% -----------------------------------------------------------------------------------------------------

function [beta,betaCov,sigma] = calc_beta(VY,X,pKX,Vmask)

n = size(VY,1);
Y = zeros([prod(VY(1).dim(1:2)) n]);
beta = zeros([VY(1).dim(1:3) size(pKX,1)]);
betaCov  = zeros([VY(1).dim(1:3) size(pKX,1)^2]);
sigma = zeros(VY(1).dim(1:3));

if nargin < 4
    Vmask = [];
elseif ~isempty(Vmask)
    mask = reslice_mask(VY(1),Vmask);
    mask = reshape(mask,VY(1).dim(1)*VY(1).dim(2),VY(1).dim(3));
end

spm_progress_bar('Init',VY(1).dim(3),'Computing','planes completed')
for j=1:VY(1).dim(3),
    
    M  = spm_matrix([0 0 j]);
    
    % Load slice j from all images
    for i=1:n
        tmp = spm_slice_vol(VY(i),M,VY(1).dim(1:2),1);
        tmp(isnan(tmp)) = 0;
        if ~isempty(Vmask)
            Y(:,i) = tmp(:).*(mask(:,j)>0);
        else
            Y(:,i) = tmp(:);
        end
    end
    pXY = pKX*Y';
    % calc betaVar
    tmpX = inv(X'*X);
    tmpX = tmpX(:);
    res_ss = diag((Y'-X*pXY)'*(Y'-X*pXY))/(n-size(X,2));
    VarpXY = (diag((Y'-X*pXY)'*(Y'-X*pXY))/(n-size(X,2))*tmpX')';
    VarpXY = reshape(VarpXY',[VY(1).dim(1:2) size(pKX,1)^2]);
    betaCov(:,:,j,:) = VarpXY;
    pXY = reshape(pXY',[VY(1).dim(1:2) size(pKX,1)]);
    beta(:,:,j,:) = pXY;
    sigma(:,:,j) = reshape(res_ss',VY(1).dim(1:2));
    spm_progress_bar('Set',j);
end

spm_progress_bar('Clear');

return

function obj = draw_slices(zscorefile,templatefile,zscorerange)
obj = slover;
%%% set parameters for template %%%%
obj.img(1).vol = spm_vol(templatefile);
[mx mn] = slover('volmaxmin',obj.img(1).vol);
obj.img(1).type = 'truecolour';
obj.img(1).cmap = gray;
obj.img(1).range = [mn mx];
%%% set parameters for zcorefile %%%%
obj.img(2).vol = spm_vol(zscorefile);
[mx mn] = slover('volmaxmin',obj.img(2).vol);
obj.img(2).type = 'split';
[cmap w]= slover('getcmap', 'hot');
obj.img(2).cmap = cmap;
obj.img(2).range = zscorerange; %%% range for zscore %%%
%%% set parameters for slover object %%%
obj.transform = eye(4);
obj.figure = spm_figure('GetWin','Graphics');
obj.cbar = 2;
obj = fill_defaults(obj);
obj.slices = -55:5:75;
params.clf = 0;
%%% do painting %%%
fy_paint(obj,params);
return
